/*
 * Copyright (c) 2016 Samsung Electronics Co., Ltd All Rights Reserved
 *
 * Licensed under the Apache License, Version 2.0 (the License);
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 * http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an AS IS BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

using System;
using System.IO;
using System.Linq;
using System.Reflection;
using System.Runtime.Loader;
using System.Collections.Generic;

namespace Tizen.Runtime.Coreclr
{
    public class AssemblyLoader : AssemblyLoadContext
    {
        private const string NativeAssemblyInfix = ".ni";
        private const string DllAssemblySuffix = ".dll";
        private const string ExeAssemblySuffix = ".exe";
        private const string NativeDllAssemblySuffix = NativeAssemblyInfix + DllAssemblySuffix;
        private static readonly string[] Suffixes = new string[] { NativeDllAssemblySuffix, DllAssemblySuffix, ExeAssemblySuffix };
        private SortedSet<string> DllDirectoriesSet = new SortedSet<string>();
        private SortedSet<string> NativeDirectoriesSet = new SortedSet<string>();
        private HashSet<FileInfo> AssemblyCacheSet = new HashSet<FileInfo>();

        public AssemblyLoader()
        {
            AssemblyLoadContext.Default.Resolving += OnResolving;
        }

        public IEnumerable<string> DllDirectories
        {
            get { return DllDirectoriesSet; }
        }

        public IEnumerable<string> NativeDirectories
        {
            get { return NativeDirectoriesSet; }
        }

        public void AddSearchableDirectory(string directory)
        {
            if (Directory.Exists(directory))
            {
                DllDirectoriesSet.Add(directory);
                NativeDirectoriesSet.Add(directory);

                foreach (var file in Directory.GetFiles(directory))
                {
                    var info = new FileInfo(file);

                    if (Suffixes.Contains(info.Extension))
                    {
                        AssemblyCacheSet.Add(info);
                    }
                }
            }
        }

        public void RemoveSearchableDirectory(string directory)
        {
            DllDirectoriesSet.Remove(directory);
            NativeDirectoriesSet.Remove(directory);

            AssemblyCacheSet.RemoveWhere(x => x.DirectoryName == directory);
        }

        public Assembly LoadFromPath(string path)
        {
            if (string.Compare(path, path.Length - NativeDllAssemblySuffix.Length, NativeAssemblyInfix, 0, NativeAssemblyInfix.Length, StringComparison.OrdinalIgnoreCase) == 0)
            {
                return LoadFromNativeImagePath(path, null);
            }
            else
            {
                return LoadFromAssemblyPath(path);
            }
        }

        protected override Assembly Load(AssemblyName assemblyName)
        {
            return Resolve(assemblyName);
        }

        protected override IntPtr LoadUnmanagedDll(string unmanagedDllName)
        {
            IntPtr native = base.LoadUnmanagedDll(unmanagedDllName);
            if (native == IntPtr.Zero)
            {
                foreach (string dir in NativeDirectories)
                {
                    FileInfo f = new FileInfo(Path.Combine(dir, unmanagedDllName));
                    if (File.Exists(f.FullName))
                    {
                        native = LoadUnmanagedDllFromPath(f.FullName);
                        break;
                    }
                }
            }

            return native;
        }

        private Assembly Resolve(AssemblyName assemblyName)
        {
            foreach (string suffix in Suffixes)
            {
                var info = AssemblyCacheSet.FirstOrDefault(x => x.Name == assemblyName.Name + suffix);

                if (info != null)
                {
                    return LoadFromPath(info.FullName);
                }
            }

            return null;
        }

        private Assembly OnResolving(AssemblyLoadContext context, AssemblyName assemblyName)
        {
            return Resolve(assemblyName);
        }
    }
}
